/***
*** read.c
***
***  General description of this file:
***     Device driver source code for General Standards 16AI32
***     family of analog input boards. This file is part of the Linux
***     driver source distribution for this board.
***
***
***  Author:
***
***  Support:
***
***/

//////////////////////////////////////////////////////////////////////////
// set the following flag to trace interrupt debug messages.
#include <windows.h>
#include <stdio.h>
#include <rtapi.h>

#include "plx_regs.h"
#include "internals.h"
#include "GSC16Ai32.h"

static int FillIntermediateBuffer(PDEVICE_BOARD device, unsigned long *buf, unsigned long nRequested, int noblock);
static int GSC16AI32_deviceReadDma(PDEVICE_BOARD device, unsigned long nRequested,int DMAMode);

/************************************************************************/
/* TransferToUser - be sure not to copy more than there is space in the */
/* user buffer.  Returns bytes transferred.                             */
/************************************************************************/
long TransferToUser(PDEVICE_BOARD device, unsigned long nRequested)
{
	int bytesRequested;

	// make sure there is space for the data.
	nRequested = min((unsigned long)device->dma_data[RX].userRemaining, nRequested);

	// make sure there is data
	nRequested = min((unsigned long)device->dma_data[RX].intermediateSamples, nRequested);

	// don't transfer more than is in the intermediate buffer.
	nRequested = min(DMA_SIZE, nRequested);

	// convert to bytes.
	bytesRequested = nRequested * sizeof(unsigned long);

	//msg("(%d): TransferToUser before userRemain %ld intermediateSamples %ld nRequested %ld\n",device->minor, device->dma_data[RX].userRemaining,device->dma_data[RX].intermediateSamples, nRequested);
	//msg("(%d): TransferToUser before userOffset %ld intermediateSamples %ldnRequested %ld\n",device->minor, (unsigned long)device->dma_data[RX].userOffset, device->dma_data[RX].intermediateSamples, nRequested);
	//msg("(%d): src base %p starting src %p dst base %p starting dst %p\n",device->minor, device->dma_data[RX].userBuffer,((char*)device->dma_data[RX].userBuffer)+device->dma_data[RX].userOffset*sizeof(u32), device->dma_data[RX].intermediateBuffer, &device->dma_data[RX].intermediateBuffer[device->dma_data[RX].intermediateStart]);

	memcpy((char*)((device->dma_data[RX].userBuffer)+(device->dma_data[RX].userOffset*sizeof(long))),
		&(device->dma_data[RX].intermediateBuffer[device->dma_data[RX].intermediateStart]), bytesRequested);

	device->dma_data[RX].intermediateStart += nRequested;
	device->dma_data[RX].intermediateSamples -= nRequested;
	device->dma_data[RX].userOffset += nRequested;
	device->dma_data[RX].userRemaining-= nRequested;
	return bytesRequested;
}

/************************************************************************/
/* Transfer data from the intermediate buffer to the user buffer.       */
/* returns bytes transferred                                            */
/************************************************************************/
int ProcessRead(PDEVICE_BOARD device)
{
	unsigned long nRequested;
	int retval;
	long bytesCopied;
	int	slept	= 0;

	nRequested=device->dma_data[RX].userRemaining;

	bytesCopied=0;

	// start by copying what is already in the intermediate buffer.


	// give either what the user requests or what is left in the buffer, whichever is less.
	if (device->dma_data[RX].intermediateSamples > 0) 
	{
		retval = TransferToUser(device, nRequested);

		// if the user buffer is now full, return.
		if (device->dma_data[RX].userRemaining==0)
		{ // done
			return(retval);
		}

		// if the user does not ask for full buffers, return
		if (!device->fillBuffer) 
		{
			return(retval);
		}
		bytesCopied+=retval;

	} // if (device->intermediateSamples > 0)

	// need to get a batch of data from the board.

	// read at most an intermediate buffer full at a time
	if (nRequested > DMA_SAMPLES) 
	{
		nRequested = DMA_SAMPLES;
	}

	do 
	{  // do it over and over if fillBuffer flag set, once otherwise
		retval = FillIntermediateBuffer(device, 
			(unsigned long *)((device->dma_data[RX].userBuffer)+(device->dma_data[RX].userOffset)),
			nRequested,
			FALSE);
		if (retval<0)
		{ // returned error
			return retval;
		}
		else if (retval)
		{
			slept	= 0;
		}
		else if (slept >= 10)
		{
			return(GSC16AI32_ERR_TRY_AGAIN);
		}
		else
		{
			slept++;
			RtSleep(1);
		}

		bytesCopied += TransferToUser(device, retval);
		nRequested = device->dma_data[RX].userRemaining;
	} 
	while((device->dma_data[RX].userRemaining>0) && device->fillBuffer);

	return (bytesCopied);
}

/************************************************************************/
/* read operation: either polled or uses PLX DMA on CH0                 */
/************************************************************************/
size_t GSC16AI32_deviceRead(PDEVICE_BOARD device, char *buf, size_t size)
{
	unsigned long nRequested;
	unsigned long fifosamples;

	nRequested = size / sizeof(unsigned long);

	// verify parameters
	if (nRequested <= 0) 
	{
		return (0);
	}


	if (device->detectOverflow) 
	{
		fifosamples = GSC16AI32_readLocal(device,BUFFER_SIZE_REG);
		if (fifosamples>=device->hwBufferSize-1)
		{ // overflow

			// clear the buffer.

			GSC16AI32_orLocal(device,INPUT_BUFF_CTRL_REG, IDC_INPUT_CLEAR_BUFFER);
			return (GSC16AI32_ERR_BUFFER_OVERFLOW);
		}
	}
	device->dma_data[RX].userOffset=0; // start at the beginning of the user buffer.
	device->dma_data[RX].userRemaining=nRequested;
	device->dma_data[RX].intermediateStart=0;
	device->dma_data[RX].userBuffer=/*(u32*)*/buf;
	return ProcessRead(device);
}

/************************************************************************/
/*  FillIntermediateBuffer                                              */
/*                                                                      */
/* Buffered I/O read method                                             */
/* The method first attempts to fill the DMA buffer to a "reasonable"   */
/* level                                                                */
/************************************************************************/
static int FillIntermediateBuffer(PDEVICE_BOARD device, unsigned long *buf, unsigned long nRequested, int noblock)
{
	int fifosamples, threshold;
	unsigned long nToTransfer, i;
	DWORD dwWaitRet;
	int nStatus = 0;
	unsigned long icr;

	nToTransfer = min(nRequested,DMA_SAMPLES);
	fifosamples=0;
	device->dma_data[RX].intermediateStart=0;
	device->dma_data[RX].intermediateSamples=0;

	switch(device->dma_data[RX].dmaState) 
	{
		// for PIO and regular DMA, wait for sufficient samples
	case DMA_DISABLE: // PIO
	case DMA_ENABLE:

		// enable the interrupt now in case the interrupt happens during the moment of indecision

		device->bLoHiPending = TRUE;
		GSC16AI32_EnableIrqLocal(device, GROUP_1, ICR_GROUP_1_LO_HI);
		icr = GSC16AI32_readLocal(device,INT_CTRL_REG);
		//printf ("icr = %08lx\n",icr);

		// check how much data there is in the FIFO
		fifosamples = GSC16AI32_readLocal(device,BUFFER_SIZE_REG);
		//threshold = GSC16AI32_readLocal(device, INPUT_BUFF_CTRL_REG)&device->hwBufferSize;
		threshold = GSC16AI32_readLocal(device, INPUT_BUFF_CTRL_REG)&0x3FFFF;
		if (fifosamples < threshold) // need more
		{
			// first of all, check for non-blocking I/O.
			if (noblock) 
			{
				return (GSC16AI32_ERR_TRY_AGAIN);
			}

			// see if there is already more data in the input data than the threshold.

			GSC16AI32_EnableIrqPlx(device);


			dwWaitRet = RtWaitForSingleObject(device->hEventBufferThresLowToHigh,device->timeout_seconds*1000);
			switch(dwWaitRet)
			{
			case WAIT_FAILED:
				device->bLoHiPending	=	FALSE;
				GSC16AI32_DisableIrqLocalAll(device);
				GSC16AI32_DisableIrqPlx(device);
				nStatus=GSC16AI32_ERR_FAILED;
				break;
			case WAIT_ABANDONED:
				device->bLoHiPending	=	FALSE;
				GSC16AI32_DisableIrqLocalAll(device);
				GSC16AI32_DisableIrqPlx(device);
				nStatus=GSC16AI32_ERR_ABANDONED;
				break;
			case WAIT_OBJECT_0:
				break;
			case WAIT_TIMEOUT:
				device->bLoHiPending	=	FALSE;
				nStatus=GSC16AI32_ERR_TIMEOUT;
				break;
			}
			if(nStatus != 0)
			{
				return nStatus;
			}

			fifosamples = GSC16AI32_readLocal(device,BUFFER_SIZE_REG);  // update samples available
			// if still no data present, something is wrong.
			//printf("fifosamples= %x\n",fifosamples);
			if (fifosamples<=0) 
			{
				return (GSC16AI32_ERR_IO);
			}
		}
		//else // don't need to wait for data...
		{
			//GSC16AI32_DisableIrqLocalAll(device);
			//device->bLoHiPending=FALSE;
		}

		break;

		nToTransfer=min(nToTransfer,(unsigned long)fifosamples);

		// for demand mode, wait for the DMA done interrupt.
	case DMA_DEMAND_MODE:
		if ((nStatus=GSC16AI32_deviceReadDma(device, nToTransfer, DMA_DEMAND_MODE))!=0)
		{
			return (nStatus);
		}
		break;

	default: // invalid state
		return (GSC16AI32_ERR_IO);
		break;
	} // switch

	// move the data from the hardware buffer to the intermediate buffer by the user's chosen method.

	if (device->dma_data[RX].dmaState != DMA_DEMAND_MODE)
	{
		nToTransfer=min(nToTransfer,(unsigned long) fifosamples);
	}

	switch(device->dma_data[RX].dmaState) 
	{
	case DMA_DISABLE: // PIO
		// don't copy more than the buffer can hold.

		for (i=0; i< nToTransfer; i++) 
		{ // PIO read
			device->dma_data[RX].intermediateBuffer[i] = GSC16AI32_readLocal(device,INPUT_DATA_BUFF_REG);
		}
		device->dma_data[RX].intermediateSamples=i;
		break;

	case DMA_ENABLE:
		if((nStatus=GSC16AI32_deviceReadDma(device, nToTransfer,DMA_ENABLE)) != 0)
		{
			return (nStatus);
		};
		break;

	case DMA_DEMAND_MODE:
		// already in the intermediate buffer.
		break;

	default: // invalid state
		return (GSC16AI32_ERR_IO);
		break;
	}


	return nToTransfer;

} // FillIntermediateBuffer

/************************************************************************/
/* device_read_dma                                                      */
/************************************************************************/
/* fill the buffer using the DMA method */
/* by the time this is called, we know that we have enough samples in the FIFO */
static int GSC16AI32_deviceReadDma(PDEVICE_BOARD device, unsigned long nRequested, int DMAMode)
{
	int nStatus = 0;
	//regdump(device);

	device->bDmaInProgress = TRUE;
	GSC16AI32_EnableIrqPlxDMA0(device);

	if (DMAMode==DMA_ENABLE)
	{
		GSC16AI32_writePlx(device, DMA_CH_0_MODE, NON_DEMAND_DMA_MODE);
	}
	else
	{
		GSC16AI32_writePlx(device, DMA_CH_0_MODE, DEMAND_DMA_MODE_READ);
	}

	GSC16AI32_writePlx(device, DMA_CH_0_PCI_ADDR, device->dma_data[RX].intermediatePhysicalAddr);
	GSC16AI32_writePlx(device, DMA_CH_0_LOCAL_ADDR, (INPUT_DATA_BUFF_REG << 2));
	GSC16AI32_writePlx(device, DMA_CH_0_TRANS_BYTE_CNT, nRequested << 2);
	GSC16AI32_writePlx(device, DMA_CH_0_DESC_PTR, /*0*/0xA);
	// OK, here we go!
	GSC16AI32_orPlx(device, DMA_CMD_STATUS, SETUP_DMA_CMD_0);
	//RtPrintf("(%d): device_read_dma - about to start dma... in buffer size: %.8X\n", device->minor,readLocal(device,BUFFER_SIZE_REG));

	//RtPrintf("(%d): device_read_dma - about to wait for dma...\n", device->minor);
	//regdump(device,"Before DMA");
	GSC16AI32_orPlx(device, DMA_CMD_STATUS, STARTUP_DMA_CMD_0);
	//RtPrintf("(%d): device_read_dma - about to wait for dma... \n", device->minor);
	nStatus=RtWaitForSingleObject(device->hEventDmaPending,device->timeout_seconds*1000);
	if (nStatus == WAIT_TIMEOUT)
	{
		return (GSC16AI32_ERR_TIMEOUT);
	}
	else if (nStatus == WAIT_OBJECT_0)
	{
		device->dma_data[RX].intermediateSamples=nRequested;
	}
	return 0;
}

